// zeal - A portable Glk-based Z-code interpreter
// Copyright (C) 2000 Jeremy Condit <jcondit@eecs.harvard.edu>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// =======================================================================
//  win.cc:
//
//  this module contains support for windows glk features.
//  right now this module is pretty minimal--it just names the app and
//  leaves the rest to the library.  note that the default osdep structure
//  will be used since we don't supply one.
// =======================================================================

#include "zeal.h"

extern "C" {
// this comes with the windowsglk library
#include "winglk.h"
}

// winglk_startup_code
//
// set the name of the app and get on with things...

int
winglk_startup_code()
{
    winglk_app_set_name("Zeal");
    winglk_window_set_title("Zeal");

    return 1;
}
