// zeal - A portable Glk-based Z-code interpreter
// Copyright (C) 2000 Jeremy Condit <jcondit@eecs.harvard.edu>
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// =======================================================================
//  zeal.h:
//
//  this header contains basic declarations used throughout zeal.  it
//  should be included by every module.
// =======================================================================

#ifndef _zeal_h_
#define _zeal_h_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern "C" {
#include "glk.h"
}

typedef unsigned int uint;

typedef uint address;

typedef unsigned short uword;
typedef signed short word;

typedef unsigned char ubyte;
typedef signed char byte;

const uint MAX_UBYTE = 255;
const uint MAX_UWORD = 65535;

#endif // _zeal_h_
